#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#
from abc import ABC
from typing import Any, Mapping

from airbyte_cdk.sources.declarative.declarative_stream import DeclarativeStream
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.migrations.state_migration import StateMigration
from airbyte_cdk.sources.types import Config


ARCHIVED_EMAIL = {"archived": "true", "campaign_type": "email"}
NOT_ARCHIVED_EMAIL = {"archived": "false", "campaign_type": "email"}

ARCHIVED = {"archived": "true"}
NOT_ARCHIVED = {"archived": "false"}

DEFAULT_START_DATE = "2012-01-01T00:00:00Z"


class ArchivedToPerPartitionStateMigration(StateMigration, ABC):
    """
    Updates old format state to new per partitioned format.
    Partitions: [{archived: True}, {archived: False}]
    Default built in airbyte cdk migration will recognise only top-level field cursor value(updated_at),
    but for partition {archived: True} source should use cursor value from archived object.

    Example input state:
    {
        "updated_at": "2020-10-10T00:00:00+00:00",
        "archived": {
          "updated_at": "2021-10-10T00:00:00+00:00"
        }
    }

    Example output state:
    {
        "partition":{ "archived":"true" },
        "cursor":{ "updated_at":"2021-10-10T00:00:00+00:00" }
    }
    {
        "partition":{ "archived":"false" },
        "cursor":{ "updated_at":"2020-10-10T00:00:00+00:00" }
    }
    """

    declarative_stream: DeclarativeStream
    config: Config

    def __init__(self, declarative_stream: DeclarativeStream, config: Config):
        self._config = config
        self.declarative_stream = declarative_stream
        self._cursor = declarative_stream.incremental_sync
        self._parameters = declarative_stream.parameters
        self._cursor_field = InterpolatedString.create(self._cursor.cursor_field, parameters=self._parameters).eval(self._config)

    def get_archived_cursor_value(self, stream_state: Mapping[str, Any]):
        return stream_state.get("archived", {}).get(self._cursor.cursor_field, self._config.get("start_date", DEFAULT_START_DATE))

    def get_not_archived_cursor_value(self, stream_state: Mapping[str, Any]):
        return stream_state.get(self._cursor.cursor_field, self._config.get("start_date", DEFAULT_START_DATE))

    def should_migrate(self, stream_state: Mapping[str, Any]) -> bool:
        return bool("states" not in stream_state and stream_state)

    def migrate(self, stream_state: Mapping[str, Any]) -> Mapping[str, Any]:
        if not self.should_migrate(stream_state):
            return stream_state
        is_archived_updated_at = self.get_archived_cursor_value(stream_state)
        is_not_archived_updated_at = self.get_not_archived_cursor_value(stream_state)

        migrated_stream_state = {
            "states": [
                {"partition": ARCHIVED, "cursor": {self._cursor.cursor_field: is_archived_updated_at}},
                {"partition": NOT_ARCHIVED, "cursor": {self._cursor.cursor_field: is_not_archived_updated_at}},
            ]
        }
        return migrated_stream_state


class CampaignsStateMigration(ArchivedToPerPartitionStateMigration):
    """
    Campaigns stream has 2 partition field: archived and campaign_type(email, sms).
    Previous API version didn't return sms in campaigns output so we need to migrate only email partition.

    Example input state:
    {
        "updated_at": "2020-10-10T00:00:00+00:00",
        "archived": {
          "updated_at": "2021-10-10T00:00:00+00:00"
        }
      }
    Example output state:
    {
        "partition":{ "archived":"true","campaign_type":"email" },
        "cursor":{ "updated_at":"2021-10-10T00:00:00+00:00" }
    }
    {
        "partition":{ "archived":"false","campaign_type":"email" },
        "cursor":{ "updated_at":"2020-10-10T00:00:00+00:00" }
    }
    """

    def migrate(self, stream_state: Mapping[str, Any]) -> Mapping[str, Any]:
        if not self.should_migrate(stream_state):
            return stream_state
        is_archived_updated_at = self.get_archived_cursor_value(stream_state)
        is_not_archived_updated_at = self.get_not_archived_cursor_value(stream_state)

        migrated_stream_state = {
            "states": [
                {"partition": ARCHIVED_EMAIL, "cursor": {self._cursor.cursor_field: is_archived_updated_at}},
                {"partition": NOT_ARCHIVED_EMAIL, "cursor": {self._cursor.cursor_field: is_not_archived_updated_at}},
            ]
        }
        return migrated_stream_state



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from typing import Any, List, Mapping, Optional

from airbyte_cdk import TState
from airbyte_cdk.models import ConfiguredAirbyteCatalog
from airbyte_cdk.sources.declarative.yaml_declarative_source import YamlDeclarativeSource


class SourceKlaviyo(YamlDeclarativeSource):
    def __init__(self, catalog: Optional[ConfiguredAirbyteCatalog], config: Optional[Mapping[str, Any]], state: TState, **kwargs):
        super().__init__(catalog=catalog, config=config, state=state, **{"path_to_yaml": "manifest.yaml"})



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

import logging
from typing import Any, Dict, Mapping, Optional

from airbyte_cdk.sources.declarative.requesters.error_handlers.backoff_strategies import WaitTimeFromHeaderBackoffStrategy
from airbyte_cdk.sources.declarative.transformations import RecordTransformation
from airbyte_cdk.sources.message.repository import InMemoryMessageRepository
from airbyte_cdk.sources.streams.call_rate import APIBudget
from airbyte_cdk.sources.streams.http.error_handlers import BackoffStrategy, ErrorHandler, HttpStatusErrorHandler
from airbyte_cdk.sources.streams.http.error_handlers.default_error_mapping import DEFAULT_ERROR_MAPPING
from airbyte_cdk.sources.streams.http.error_handlers.response_models import ErrorResolution, FailureType, ResponseAction
from airbyte_cdk.sources.streams.http.http_client import HttpClient
from airbyte_cdk.sources.types import Config, StreamSlice, StreamState


class CampaignsDetailedTransformation(RecordTransformation):
    """
    Campaigns detailed stream fetches detailed campaigns info:
    estimated_recipient_count: integer
    campaign_messages: list of objects.

    To get this data CampaignsDetailedTransformation makes extra API requests:
    https://a.klaviyo.com/api/campaign-recipient-estimations/{campaign_id}
    https://developers.klaviyo.com/en/v2024-10-15/reference/get_messages_for_campaign
    """

    config: Config

    api_revision = "2024-10-15"
    url_base = "https://a.klaviyo.com/api/"
    name = "campaigns_detailed"
    max_retries = 5
    max_time = 60 * 10

    def __init__(self, config: Config, **kwargs):
        self.logger = logging.getLogger("airbyte")
        self.config = config
        self._api_key = self.config["api_key"]
        self._http_client = HttpClient(
            name=self.name,
            logger=self.logger,
            error_handler=self.get_error_handler(),
            api_budget=APIBudget(policies=[]),
            backoff_strategy=self.get_backoff_strategy(),
            message_repository=InMemoryMessageRepository(),
        )

    def transform(
        self,
        record: Dict[str, Any],
        config: Optional[Config] = None,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
    ) -> None:
        self._set_recipient_count(record)
        self._set_campaign_message(record)

    def _set_recipient_count(self, record: Mapping[str, Any]) -> None:
        campaign_id = record["id"]
        _, recipient_count_response = self._http_client.send_request(
            url=f"{self.url_base}campaign-recipient-estimations/{campaign_id}",
            request_kwargs={},
            headers=self.request_headers(),
            http_method="GET",
        )
        record["estimated_recipient_count"] = (
            recipient_count_response.json().get("data", {}).get("attributes", {}).get("estimated_recipient_count", 0)
        )

    def _set_campaign_message(self, record: Mapping[str, Any]) -> None:
        messages_link = record.get("relationships", {}).get("campaign-messages", {}).get("links", {}).get("related")
        if messages_link:
            _, campaign_message_response = self._http_client.send_request(
                url=messages_link, request_kwargs={}, headers=self.request_headers(), http_method="GET"
            )
            record["campaign_messages"] = campaign_message_response.json().get("data")

    def get_backoff_strategy(self) -> BackoffStrategy:
        return WaitTimeFromHeaderBackoffStrategy(header="Retry-After", max_waiting_time_in_seconds=self.max_time, parameters={}, config={})

    def request_headers(self):
        return {
            "Accept": "application/json",
            "Revision": self.api_revision,
            "Authorization": f"Klaviyo-API-Key {self._api_key}",
        }

    def get_error_handler(self) -> ErrorHandler:
        error_mapping = DEFAULT_ERROR_MAPPING | {
            404: ErrorResolution(ResponseAction.IGNORE, FailureType.config_error, "Resource not found. Ignoring.")
        }

        return HttpStatusErrorHandler(logger=self.logger, error_mapping=error_mapping, max_retries=self.max_retries)
